import multiprocessing
import queue
import random
import string
from multiprocessing import Queue

from .errors import AbortError


class ResultProcess(multiprocessing.Process):
    def __init__(self, target, **kwargs) -> None:
        self._real_target = target
        self._result_queue: Queue = Queue()
        self._failed = multiprocessing.Event()
        kwargs.setdefault("daemon", True)
        super(ResultProcess, self).__init__(target=self._wrapper, **kwargs)

    def _wrapper(self, *args, **kwargs) -> None:
        try:
            # print(args, kwargs)
            self._result_queue.put(self._real_target(*args, **kwargs))
        except BaseException as ex:
            self._failed.set()
            self._result_queue.put(ex)

    def abort(self) -> None:
        self._failed.set()
        self._result_queue.put(AbortError())

    def get_result(self, timeout: float = None):
        result = None
        try:
            result = self._result_queue.get(timeout=timeout)
        except queue.Empty:
            raise TimeoutError("Operation timed out.")
        if self._failed.is_set():
            raise result
        return result


def generate_token(length: int) -> str:
    chars = string.ascii_letters + string.digits
    return "".join(random.choice(chars) for _ in range(length))



import argparse
import configparser
import logging
import os
import time
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List
from urllib.parse import quote_plus, unquote_plus

import multivolumefile
import py7zr

from . import __version__
from .client import ToDusClient

logging.basicConfig(format="%(levelname)s - %(message)s", level=logging.INFO)


client = ToDusClient()
config = configparser.ConfigParser()


def write_txt(filename: str, urls: List[str], parts: List[str]) -> str:
    txt = "\n".join(f"{down_url}\t{name}" for down_url, name in zip(urls, parts))
    path = Path(f"{filename}.txt").resolve()
    with open(path, "w", encoding="utf-8") as f:
        f.write(txt)
    return str(path)


def split_upload(token: str, path: str, part_size: int) -> str:
    global client

    with open(path, "rb") as file:
        data = file.read()
    filename = os.path.basename(path)
    with TemporaryDirectory() as tempdir:
        with multivolumefile.open(
            os.path.join(tempdir, filename + ".7z"),
            "wb",
            volume=part_size,
        ) as vol:
            with py7zr.SevenZipFile(vol, "w") as a:  # type: ignore
                a.writestr(data, filename)
        del data
        parts = sorted(os.listdir(tempdir))
        parts_count = len(parts)

        urls = []

        for i, name in enumerate(parts, 1):
            logging.info(f"Uploading {i}/{parts_count}: {filename}")
            with open(os.path.join(tempdir, name), "rb") as file:
                part = file.read()
            try:
                urls.append(client.upload_file(token, part, len(part)))
            except Exception as ex:
                logging.exception(ex)
                time.sleep(15)
                try:
                    urls.append(client.upload_file(token, part, len(part)))
                except Exception as ex:
                    logging.exception(ex)
                    raise ValueError(
                        f"Failed to upload part {i} ({len(part):,}B): {ex}"
                    )
        path = write_txt(filename, urls, parts)
        return path


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog=__name__.split(".")[0],
        description="ToDus Client",
    )
    parser.add_argument(
        "-n",
        "--number",
        dest="number",
        metavar="PHONE-NUMBER",
        help="account's phone number",
        required=True,
    )
    parser.add_argument(
        "-c",
        "--config-folder",
        dest="folder",
        type=str,
        default=".",
        help="folder where account configuration will be saved/loaded",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=__version__,
        help="show program's version number and exit.",
    )

    subparsers = parser.add_subparsers(dest="command")

    login_parser = subparsers.add_parser(name="login", help="authenticate in server")

    up_parser = subparsers.add_parser(name="upload", help="upload file")
    up_parser.add_argument(
        "-p",
        "--part-size",
        dest="part_size",
        type=int,
        default=0,
        help="if given, the file will be split in parts of the given size in bytes",
    )
    up_parser.add_argument("file", nargs="+", help="file to upload")

    down_parser = subparsers.add_parser(name="download", help="download file")
    down_parser.add_argument("url", nargs="+", help="url to download or txt file path")

    return parser


def register(client: ToDusClient, phone: str) -> str:
    client.request_code(phone)
    pin = input("Enter PIN:").strip()
    password = client.validate_code(phone, pin)
    logging.debug("PASSWORD: %s", password)
    return password


def get_password(phone: str, folder: str) -> str:
    config_path = Path(folder) / Path(f"{phone}.ini")
    config.read(config_path)
    return config["DEFAULT"].get("password", "") if "DEFAULT" in config else ""


def save_config(phone: str, folder: str = ".") -> None:
    with open(Path(folder) / Path(f"{phone}.ini"), "w") as configfile:
        config.write(configfile)


def main() -> None:
    global client

    parser = get_parser()
    args = parser.parse_args()
    phone = args.number
    password = get_password(phone, args.folder) or "***FIX_TEMP***"

    if not password and args.command != "login":
        print("ERROR: account not authenticated, login first.")
        return

    if args.command == "upload":
        # token = client.login(phone, password)
        token = config["DEFAULT"].get("token", "") if "DEFAULT" in config else ""
        logging.debug(f"Token: '{token}'")

        for path in args.file:
            filename = os.path.basename(path)
            logging.info(f"Uploading: {filename}")
            if args.part_size:
                txt = split_upload(token, path, args.part_size)
                logging.info(f"TXT: {txt}")
            else:
                with open(path, "rb") as file:
                    data = file.read()
                url = client.upload_file(token, data, len(data))
                down_url = f"{url}?name={quote_plus(filename)}"
                logging.info(f"URL: {down_url}")
                txt = write_txt(filename, urls=[url], parts=[filename])
                logging.info(f"TXT: {txt}")
    elif args.command == "download":
        # token = client.login(phone, password)
        token = config["DEFAULT"].get("token", "") if "DEFAULT" in config else ""
        logging.debug(f"Token: '{token}'")

        while args.url:
            url = args.url.pop(0)
            if os.path.exists(url):
                with open(url) as fp:
                    urls = []
                    for line in fp.readlines():
                        line = line.strip()
                        if line:
                            urls.append("{}?name={}".format(*line.split(maxsplit=1)))

                    args.url = urls + args.url
                    continue
            logging.info(
                f"Downloading: {url}",
            )
            url, name = url.split("?name=", maxsplit=1)
            name = unquote_plus(name)
            try:
                size = client.download_file(token, url, name)
            except Exception:
                size = client.download_file(token, url, name)
            logging.debug(
                f"File Size: {size // 1024}",
            )
    elif args.command == "login":
        # TODO: Fix Register
        # password = register(client, phone)
        # token = client.login(phone, password)

        token = config["DEFAULT"].get("token", "") if "DEFAULT" in config else ""
        logging.debug(f"Token: '{token}'")

        config["DEFAULT"]["password"] = password
        config["DEFAULT"]["token"] = token
        save_config(phone, args.folder)
    else:
        parser.print_usage()



import json
import re
import socket
import ssl
from base64 import b64decode, b64encode
from typing import Tuple

from .errors import AuthenticationError
from .util import generate_token


def _get_socket() -> ssl.SSLSocket:
    sock = socket.socket(socket.AF_INET)
    sock.settimeout(15)
    so = ssl.wrap_socket(sock, ssl_version=ssl.PROTOCOL_TLSv1_2)
    so.connect(("im.todus.cu", 1756))
    so.send(b"<stream:stream xmlns='jc' o='im.todus.cu' xmlns:stream='x1' v='1.0'>")
    return so


def _negociate_start(m: str, so: ssl.SSLSocket, authstr: bytes, sid: str) -> bool:
    if m.startswith("<?xml version='1.0'?><stream:stream i='") and m.endswith(
        "xmlns:stream='x1' f='im.todus.cu' xmlns='jc'>"
    ):
        return True

    if m == (
        "<stream:features>"
        "<es xmlns='x2'><e>PLAIN</e><e>X-OAUTH2</e></es>"
        "<register xmlns='http://jabber.org/features/iq-register'/>"
        "</stream:features>"
    ):
        so.send(b"<ah xmlns='ah:ns' e='PLAIN'>" + authstr + b"</ah>")
        return True

    if m == "<ok xmlns='x2'/>":
        so.send(b"<stream:stream xmlns='jc' o='im.todus.cu' xmlns:stream='x1' v='1.0'>")
        return True

    if "<stream:features><b1 xmlns='x4'/>" in m:
        so.send("<iq i='{}-1' t='set'><b1 xmlns='x4'></b1></iq>".format(sid).encode())
        return True

    return False


def _parse_token(token: str) -> Tuple[str, bytes]:
    phone = json.loads(b64decode(token.split(".")[1]).decode())["username"]
    authstr = b64encode((chr(0) + phone + chr(0) + token).encode("utf-8"))
    return phone, authstr


def reserve_url(token: str, filesize: int) -> Tuple[str, str]:
    """Reserve file URL to upload.

    Returns a tuple with upload and download URLs.
    """
    phone, authstr = _parse_token(token)
    sid = generate_token(5)
    so = _get_socket()

    while True:
        m = so.recv().decode()
        if _negociate_start(m, so, authstr, sid):
            continue

        if "t='result' i='{}-1'>".format(sid) in m:
            so.send(b"<en xmlns='x7' u='true' max='300'/>")
            so.send(
                (
                    "<iq i='"
                    + sid
                    + "-3' t='get'><query xmlns='todus:purl' type='0' persistent='false' size='"
                    + str(filesize)
                    + "' room=''></query></iq>"
                ).encode()
            )
            continue

        if m.startswith("<ed u='true' max='300'"):
            so.send(("<p i='" + sid + "-4'></p>").encode())
            continue

        if m.startswith("<iq o='" + phone + "@im.todus.cu"):
            up_url = down_url = ""
            match = re.match(r".*put='(.*)' get='(.*)' stat.*", m)
            if match:
                up_url = match.group(1).replace("amp;", "")
                down_url = match.group(2)
            return (up_url, down_url)

        if "<not-authorized/>" in m:
            raise AuthenticationError()


def get_real_url(token: str, url: str) -> str:
    """Get authenticated URL."""
    phone, authstr = _parse_token(token)
    sid = generate_token(5)
    so = _get_socket()

    while True:
        m = so.recv().decode()
        if _negociate_start(m, so, authstr, sid):
            continue

        if "t='result' i='{}-1'>".format(sid) in m:
            so.send(
                "<iq i='{}-2' t='get'><query xmlns='todus:gurl' url='{}'></query></iq>".format(
                    sid, url
                ).encode(),
            )
            continue

        if "t='result' i='{}-2'>".format(sid) in m and "status='200'" in m:
            down_url = ""
            match = re.match(".*du='(.*)' stat.*", m)
            if match:
                down_url = match.group(1).replace("amp;", "")
            return down_url

        if "<not-authorized/>" in m:
            raise AuthenticationError()



import string
from typing import Any, Callable, Dict, Optional, Tuple

import requests

from .s3 import get_real_url, reserve_url
from .util import ResultProcess, generate_token


class ToDusClient:
    def __init__(
        self, version_name: str = "0.40.16", version_code: str = "21820"
    ) -> None:
        self.version_name = version_name
        self.version_code = version_code

        self.timeout = 60
        self.session = requests.Session()
        self.session.headers.update(
            {
                "Accept-Encoding": "gzip",
            }
        )
        self._real_request = self.session.request
        self.session.request = self._request  # type: ignore[assignment]
        self._process: Optional[ResultProcess] = None

    def _request(self, *args, **kwargs) -> requests.Response:
        kwargs.setdefault("timeout", self.timeout)
        return self._real_request(*args, **kwargs)

    def _run_task(self, task: Callable, timeout: float, **kwargs) -> Any:
        self._process = ResultProcess(target=task, **kwargs)
        try:
            self._process.start()
            return self._process.get_result(timeout)
        except (AttributeError, PermissionError) as ex:
            # AttributeError: Can't pickle local object 'ToDusClient.login.<locals>.task'
            raise ex
        finally:
            self.abort()

    def abort(self) -> None:
        if self._process is not None and self._process.is_alive():
            self._process.kill()  # type: ignore [attr-defined]
            self._process.abort()
            self._process = None

    @property
    def auth_ua(self) -> str:
        return f"ToDus {self.version_name} Auth"

    @property
    def upload_ua(self) -> str:
        return f"ToDus {self.version_name} HTTP-Upload"

    @property
    def download_ua(self) -> str:
        return f"ToDus {self.version_name} HTTP-Download"

    @property
    def headers_auth(self) -> Dict[str, str]:
        return {
            "Host": "auth.todus.cu",
            "User-Agent": self.auth_ua,
            "Content-Type": "application/x-protobuf",
        }

    def task_request_code(self, phone_number: str) -> None:
        # TODO: Fix payload/data
        headers = self.headers_auth
        data = (
            b"\n\n"
            + phone_number.encode("utf-8")
            + b"\x12\x96\x01"
            + generate_token(150).encode("utf-8")
        )
        url = "https://auth.todus.cu/v2/auth/users.reserve"
        with self.session.post(url, data=data, headers=headers) as resp:
            resp.raise_for_status()

    def request_code(self, phone_number: str) -> None:
        """Request server to send verification SMS code."""
        kwargs = {"args": (phone_number,)}
        self._run_task(self.task_request_code, self.timeout, **kwargs)

    def task_validate_code(self, phone_number: str, code: str) -> str:
        # TODO: Fix payload/data
        headers = self.headers_auth
        data = (
            b"\n\n"
            + phone_number.encode("utf-8")
            + b"\x12\x96\x01"
            + generate_token(150).encode("utf-8")
            + b"\x1a\x06"
            + code.encode()
        )
        url = "https://auth.todus.cu/v2/auth/users.register"
        with self.session.post(url, data=data, headers=headers) as resp:
            resp.raise_for_status()
            if b"`" in resp.content:
                index = resp.content.index(b"`") + 1
                return resp.content[index : index + 96].decode()
            else:
                return resp.content[5:166].decode()

    def validate_code(self, phone_number: str, code: str) -> str:
        """Validate phone number with received SMS code.

        Returns the account password.
        """
        kwargs = {"args": (phone_number, code)}
        return self._run_task(self.task_validate_code, self.timeout, **kwargs)

    def task_login(self, phone_number: str, password: str) -> str:
        # TODO: Fix payload/data
        headers = self.headers_auth
        data = (
            b"\n\n"
            + phone_number.encode()
            + b"\x12\x96\x01"
            + generate_token(150).encode("utf-8")
            + b"\x12\x60"
            + password.encode()
            + b"\x1a\x05"
            + self.version_code.encode("utf-8")
        )
        url = "https://auth.todus.cu/v2/auth/token"
        with self.session.post(url, data=data, headers=headers) as resp:
            resp.raise_for_status()
            token = "".join([c for c in resp.text if c in string.printable])
            return token

    def login(self, phone_number: str, password: str) -> str:
        """Login with phone number and password to get an access token."""
        kwargs = {"args": (phone_number, password)}
        return self._run_task(self.task_login, self.timeout, **kwargs)

    def task_upload_file_1(self, token: str, size: int) -> Tuple[str, str]:
        return reserve_url(token, size)

    def task_upload_file_2(
        self, token: str, data: bytes, up_url: str, down_url: str, timeout: float
    ) -> str:
        headers = {
            "User-Agent": self.upload_ua,
            "Authorization": f"Bearer {token}",
        }
        with self.session.put(
            url=up_url, data=data, headers=headers, timeout=timeout
        ) as resp:
            resp.raise_for_status()
        return down_url

    def upload_file(self, token: str, data: bytes, size: int = None) -> str:
        """Upload data and return the download URL."""
        if size is None:
            size = len(data)

        kwargs = {"args": (token, size)}
        up_url, down_url = self._run_task(
            self.task_upload_file_1, self.timeout, **kwargs
        )

        timeout = max(len(data) / 1024 / 1024 * 20, self.timeout)

        kwargs = {"args": (token, data, up_url, down_url, timeout)}  # type: ignore [dict-item]
        return self._run_task(self.task_upload_file_2, timeout, **kwargs)

    def task_download_1(self, token, url) -> str:
        return get_real_url(token, url)

    def task_download_2(self, token, url, path) -> int:
        headers = {
            "User-Agent": self.download_ua,
            "Authorization": f"Bearer {token}",
        }
        with self.session.get(url=url, headers=headers) as resp:
            resp.raise_for_status()
            size = int(resp.headers.get("Content-Length", 0))
            with open(path, "wb") as file:
                file.write(resp.content)
            return size

    def download_file(self, token: str, url: str, path: str) -> int:
        """Download file URL.

        Returns the file size.
        """
        kwargs = {"args": (token, url)}
        url = self._run_task(self.task_download_1, self.timeout, **kwargs)
        kwargs = {"args": (token, url, path)}  # type: ignore [dict-item]
        return self._run_task(self.task_download_2, self.timeout, **kwargs)



"""Setup module installation."""

import os
import re

from setuptools import find_packages, setup


def get_requirements(file_path):
    with open(file_path, encoding="utf-8") as req:
        requires_deps = [
            line.replace("==", ">=")
            for line in req.read().split("\n")
            if line and not line.startswith(("#", "-"))
        ]
    return requires_deps


if __name__ == "__main__":
    MODULE_NAME = "todus3"
    DESC = "ToDus client"
    URL = "https://github.com/oleksis/todus/tree/todus3"
    version = ""

    with open(os.path.join(MODULE_NAME, "__init__.py"), encoding="utf-8") as fh:
        m = re.search(r"__version__ = \"(.*?)\"", fh.read(), re.M)
        if m:
            version = m.group(1)

    with open("README.md", encoding="utf-8") as fh:
        long_description = fh.read()

    install_requires = get_requirements("requirements/requirements.txt")
    test_deps = get_requirements("requirements/requirements-test.txt")
    dev_deps = get_requirements("requirements/requirements-dev.txt")

    setup(
        name=MODULE_NAME,
        version=version,
        description=DESC,
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="adbenitez",
        author_email="adbenitez@nauta.cu",
        maintainer="Oleksis Fraga",
        maintainer_email="oleksis.fraga@gmail.com",
        url=URL,
        keywords="todus,s3,client",
        license="MPL",
        classifiers=[
            "Development Status :: 4 - Beta",
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
            "Operating System :: OS Independent",
            "Topic :: Utilities",
        ],
        zip_safe=False,
        include_package_data=True,
        packages=find_packages(),
        install_requires=install_requires,
        extras_require={
            "test": test_deps,
            "dev": dev_deps,
        },
        python_requires=">=3.6.0",
        entry_points={"console_scripts": [f"{MODULE_NAME} = {MODULE_NAME}.main:main"]},
    )
